//	window.cpp - The main window for the application.
//	----------------------------------------------------------------------------
//	This file is part of Mouse To OSC, a simple OSC-outputting program.
//	Copyright (c) 2006 Niall Moody.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//	----------------------------------------------------------------------------

#include <juce.h>
#include "window.h"
#include "PadControl.h"
#include "SliderControl.h"
#include "FileLoader.h"
#include "AboutBox.h"
#include "OSCMessage.h"
#include "PhilChooser.h"
#include "PhilLookAndFeel.h"
#include <iostream>

using namespace std;

//------------------------------------------------------------------------------
WindowContents::WindowContents():
Component(),
lastSlot(0),
midiInputDevice(0)
{
	int i;
	StringArray midiInputs = MidiInput::getDevices();

	//IP address to send to.
	addAndMakeVisible(addressLabel = new Label(T("aL"), T("IP Address:")));
	addAndMakeVisible(addressEditor = new TextEditor());
	addressEditor->setText(T("127.0.0.1"));
	//Port to send to.
	addAndMakeVisible(portLabel = new Label(T("pL"), T("Port:")));
	addAndMakeVisible(portEditor = new TextEditor());
	portEditor->setText(T("5678"));
	//MIDI input to receive from.
	addAndMakeVisible(midiInputLabel = new Label(T("mL"), T("MIDI Input:")));
	addAndMakeVisible(midiInput = new ComboBox(T("MIDI Input")));
	for(i=0;i<midiInputs.size();++i)
		midiInput->addItem(midiInputs[i], (i+1));
	if(midiInputs.size() > 0)
	{
		midiInput->setSelectedId(MidiInput::getDefaultDeviceIndex()+1);
		midiInputDevice = MidiInput::openDevice(MidiInput::getDefaultDeviceIndex(),
												this);
		if(midiInputDevice)
			midiInputDevice->start();
	}
	midiInput->addListener(this);

	//The 3 pads.
	pads = new PadControl *[3];
	for(i=0;i<3;++i)
	{
		addAndMakeVisible(pads[i] = new PadControl());
		pads[i]->addChangeListener(this);
	}

	//The 1st 3 sliders.
	topSliders = new SliderControl *[3];
	for(i=0;i<3;++i)
	{
		addAndMakeVisible(topSliders[i] = new SliderControl());
		topSliders[i]->addChangeListener(this);
	}

	//The 2nd 3 sliders.
	bottomSliders = new SliderControl *[3];
	for(i=0;i<3;++i)
	{
		addAndMakeVisible(bottomSliders[i] = new SliderControl());
		bottomSliders[i]->addChangeListener(this);
	}

	//The 8 'slot' buttons.
	slotButtons = new TextButton *[8];
	for(i=0;i<8;++i)
	{
		String tempstr;
		tempstr << T("Slot ") << (i+1);
		addAndMakeVisible(slotButtons[i] = new TextButton(tempstr));
		slotButtons[i]->setClickingTogglesState(true);
        slotButtons[i]->setRadioGroupId(SlotID);
		slotButtons[i]->setConnectedEdges(((i != 0) ? Button::ConnectedOnLeft : 0) |
                               			  ((i != 7) ? Button::ConnectedOnRight : 0));
		slotButtons[i]->setBackgroundColours(Colours::lightblue.brighter(0.25f),
											 Colours::cornflowerblue);
		slotButtons[i]->addButtonListener(this);
        if(i == lastSlot)
            slotButtons[i]->setToggleState (true, false);
	}
}

//------------------------------------------------------------------------------
WindowContents::~WindowContents()
{
	//Stop midiInputDevice.
	if(midiInputDevice)
	{
		midiInputDevice->stop();
		delete midiInputDevice;
	}

	//Clear up SocketSetup singleton.
	delete SocketSetup::getInstance();

	deleteAllChildren();
	delete [] pads;
	delete [] topSliders;
	delete [] bottomSliders;
}

//------------------------------------------------------------------------------
void WindowContents::resized()
{
	int i;
	int tempTop;

	if(getWidth() < 300)
	{
		addressLabel->setBounds(2, 4, (getWidth()/6), 22);
		addressEditor->setBounds((getWidth()/6)+4, 4, (getWidth()/2), 22);
		portLabel->setBounds((2*getWidth()/3)+6, 4, (getWidth()/6), 22);
		portEditor->setBounds((5*getWidth()/6)+8, 4, (getWidth()/6), 22);
	}
	else
	{
		addressLabel->setBounds(2, 4, 75, 22);
		addressEditor->setBounds(79, 4, 150, 22);
		portLabel->setBounds(240, 4, 35, 22);
		portEditor->setBounds(277, 4, 50, 22);
		midiInputLabel->setBounds(340, 4, 75, 22);
		midiInput->setBounds(418, 4, 175, 22);
	}

	for(i=0;i<8;++i)
	{
		const int tempWidth = 400;

		slotButtons[i]->setBounds(((getWidth()-tempWidth)/2)+(i*(tempWidth/8)),
								  36,
								  (tempWidth/8),
								  24);
	}

	//tempTop = 44;
	tempTop = 76;
	pads[0]->setBounds(9, tempTop, ((getWidth()/3)-32), (getWidth()/3));
	pads[1]->setBounds(((getWidth()/3)+5),
					  tempTop,
					  ((getWidth()/3)-32),
					  (getWidth()/3));
	pads[2]->setBounds(((2*getWidth())/3)+7,
					  tempTop,
					  ((getWidth()/3)-32),
					  (getWidth()/3));

	tempTop += ((getWidth()/3)+10);
	topSliders[0]->setBounds(4, tempTop, ((getWidth()/3)-4), 70);
	topSliders[1]->setBounds(((getWidth()/3)+2), tempTop, ((getWidth()/3)-4), 70);
	topSliders[2]->setBounds(((2*getWidth())/3), tempTop, ((getWidth()/3)-4), 70);

	tempTop += 84;
	bottomSliders[0]->setBounds(4, tempTop, ((getWidth()/3)-4), 70);
	bottomSliders[1]->setBounds(((getWidth()/3)+2), tempTop, ((getWidth()/3)-4), 70);
	bottomSliders[2]->setBounds(((2*getWidth())/3), tempTop, ((getWidth()/3)-4), 70);
}

//------------------------------------------------------------------------------
void WindowContents::paint(Graphics& g)
{
	g.setColour(Colours::black.withAlpha(0.5f));
	g.drawLine(3, 30, (getWidth()-4), 30);
	g.drawLine(3, 66, (getWidth()-4), 66);
	g.drawLine(3,
			   (((getWidth()/3)-28)+108),
			   (getWidth()-4),
			   (((getWidth()/3)-28)+108));
	g.drawLine(3,
			   (((getWidth()/3)-28)+108+84),
			   (getWidth()-4),
			   (((getWidth()/3)-28)+108+84));
}

//------------------------------------------------------------------------------
void WindowContents::changeListenerCallback(void *objectThatHasChanged)
{
	OSC::Message message;
	OSC::Message *bundleMessage;
	long dataSize;

	if((addressEditor->getText() == T("")) || (portEditor->getText() == T("")))
		return;

	sock.setAddress(static_cast<const char *>(addressEditor->getText()));
	sock.setPort(portEditor->getText().getIntValue());

	if((objectThatHasChanged == pads[0])||
	   (objectThatHasChanged == pads[1])||
	   (objectThatHasChanged == pads[2]))
	{
		PadControl *tempPad = static_cast<PadControl *>(objectThatHasChanged);

		//Add x message.
		bundleMessage = new OSC::Message();
		bundleMessage->setAddress(static_cast<const char *>(tempPad->getXAddress()));
		if(bundleMessage->getAddress() != "")
			bundleMessage->addFloat32(tempPad->getXVal());
		bundle.addElement(bundleMessage);

		//Add y message.
		bundleMessage = new OSC::Message();
		bundleMessage->setAddress(static_cast<const char *>(tempPad->getYAddress()));
		if(bundleMessage->getAddress() != "")
			bundleMessage->addFloat32(tempPad->getYVal());
		bundle.addElement(bundleMessage);

		//Send bundle.
		dataSize = bundle.getSize();
		sock.sendData(bundle.getData(), dataSize);

		//Clean up for the next send.
		bundle.clearElements();
	}
	else if((objectThatHasChanged == topSliders[0])||
			(objectThatHasChanged == topSliders[1])||
			(objectThatHasChanged == topSliders[2])||
			(objectThatHasChanged == bottomSliders[0])||
			(objectThatHasChanged == bottomSliders[1])||
			(objectThatHasChanged == bottomSliders[2]))
	{
		SliderControl *tempSlider = static_cast<SliderControl *>(objectThatHasChanged);

		//Send slider message.
		message.setAddress(static_cast<const char *>(tempSlider->getAddress()));
		if(message.getAddress() != "")
		{
			message.addFloat32(tempSlider->getSliderVal());
			dataSize = message.getSize();
			sock.sendData(message.getData(), dataSize);
		}
	}
}

//------------------------------------------------------------------------------
void WindowContents::buttonClicked(Button *button)
{
	int i;

	if(button->getRadioGroupId() == SlotID)
	{
		fillSlotValues(slotValues[lastSlot]);
		for(i=0;i<8;++i)
		{
			if(slotButtons[i] == button)
			{
				lastSlot = i;
				break;
			}
		}
		updateWidgets(slotValues[lastSlot]);
	}
}

//------------------------------------------------------------------------------
void WindowContents::comboBoxChanged(ComboBox *comboBoxThatHasChanged)
{
	if(midiInputDevice)
	{
		midiInputDevice->stop();
		delete midiInputDevice;
	}

	midiInputDevice = MidiInput::openDevice(midiInput->getSelectedId()-1, this);
	if(midiInputDevice)
		midiInputDevice->start();
}

//------------------------------------------------------------------------------
const StringArray WindowContents::getMenuBarNames(MenuBarComponent *menuBar)
{
	const char *names[] = {"File",
						   "Help"};
	const StringArray retval(names, 2);

	return retval;
}

//------------------------------------------------------------------------------
const PopupMenu WindowContents::getMenuForIndex(MenuBarComponent *menuBar,
												int topLevelMenuIndex,
												const String &menuName)
{
	PopupMenu retval;

	switch(topLevelMenuIndex)
	{
		case 0:
			retval.addItem(FileNew, T("New"));
			retval.addSeparator();
			retval.addItem(FileOpen, T("Open..."));
			retval.addItem(FileSave, T("Save"));
			retval.addItem(FileSaveAs, T("Save As..."));
			retval.addSeparator();
			retval.addItem(FileQuit, T("Quit"));
			break;
		case 1:
			retval.addItem(HelpAbout, T("About"));
			break;
	}

	return retval;
}
//------------------------------------------------------------------------------
void WindowContents::menuItemSelected(MenuBarComponent *menuBar,
						  			  int menuItemID,
						 			  int topLevelMenuIndex)
{
	switch(menuItemID)
	{
		case FileNew:
			fileNew();
			break;
		case FileOpen:
			fileOpen();
			break;
		case FileSave:
			fileSave();
			break;
		case FileSaveAs:
			fileSaveAs();
			break;
		case FileQuit:
			JUCEApplication::quit();
			break;
		case HelpAbout:
			{
				AboutBox dlg;
				dlg.setSize(300, 225);

				DialogWindow::showModalDialog(T("About"),
											  &dlg,
											  0,
											  Colours::lightblue,
											  true);
			}
			break;
	}
}

//------------------------------------------------------------------------------
void WindowContents::handleIncomingMidiMessage(MidiInput *source,
								   			   const MidiMessage &message)
{
	int i;
	float tempf;
	String tempstr;

	//Just in case...
	if(source == midiInputDevice)
	{
		if(message.isController())
		{
			tempstr << message.getControllerNumber();

			for(i=0;i<3;++i)
			{
				if(pads[i]->getXCc() == tempstr)
				{
					tempf = (float)message.getControllerValue()/127.0f;
					pads[i]->setXValNorm(tempf);
				}
				if(pads[i]->getYCc() == tempstr)
				{
					tempf = (float)message.getControllerValue()/127.0f;
					pads[i]->setYValNorm(tempf);
				}
				if(topSliders[i]->getCc() == tempstr)
				{
					tempf = (float)message.getControllerValue()/127.0f;
					topSliders[i]->setValueNorm(tempf);
				}
				if(bottomSliders[i]->getCc() == tempstr)
				{
					tempf = (float)message.getControllerValue()/127.0f;
					bottomSliders[i]->setValueNorm(tempf);
				}
			}
		}
	}
}

//------------------------------------------------------------------------------
void WindowContents::fileNew()
{
	int i;
	SlotValues tempValues;

	//tempValues is initialised to the correct values, so we can just update the
	//widgets from it.
	updateWidgets(tempValues);

	//Now we update the slots from the now-initialised widgets.
	for(i=0;i<8;++i)
		fillSlotValues(slotValues[i]);

	lastFile = String::empty;
}

//------------------------------------------------------------------------------
void WindowContents::fileOpen()
{
	String tempIp;
	String tempPort;
	int tempMidi;
	PhilLookAndFeel tempLaf;
	PhilChooser dlg(T("Open file..."),
					lastDir,
					T("*.xml"));

	dlg.setLookAndFeel(&tempLaf);
	dlg.setColours(Colours::lightblue,
				   Colours::lightblue.darker(0.5f).withSaturation(0.25f),
				   Colours::lightblue.darker(0.5f).withSaturation(0.75f),
				   Colours::lightblue.brighter(1.0f));
	if(dlg.browseForFileToOpen())
	{
		FileLoader fil(dlg.getResult().getFullPathName());

		fil.loadFile(slotValues, tempIp, tempPort, tempMidi);

		addressEditor->setText(tempIp);
		portEditor->setText(tempPort);
		midiInput->setSelectedId(tempMidi);

		updateWidgets(slotValues[lastSlot]);
		lastFile = dlg.getResult().getFullPathName();
		lastDir = dlg.getResult().getParentDirectory();
	}
}

//------------------------------------------------------------------------------
void WindowContents::fileSave()
{
	if(lastFile == String::empty)
		fileSaveAs();
	else
	{
		FileLoader fil(lastFile);

		fillSlotValues(slotValues[lastSlot]);
		fil.saveFile(slotValues,
					 addressEditor->getText(),
					 portEditor->getText(),
					 midiInput->getSelectedId());
	}
}

//------------------------------------------------------------------------------
void WindowContents::fileSaveAs()
{
	PhilLookAndFeel tempLaf;
	PhilChooser dlg(T("Save file as..."),
					lastDir,
					T("*.xml"));

	dlg.setLookAndFeel(&tempLaf);
	dlg.setColours(Colours::lightblue,
				   Colours::lightblue.darker(0.5f).withSaturation(0.25f),
				   Colours::lightblue.darker(0.5f).withSaturation(0.75f),
				   Colours::lightblue.brighter(1.0f));
	if(dlg.browseForFileToSave())
	{
		FileLoader fil(dlg.getResult().getFullPathName());

		fillSlotValues(slotValues[lastSlot]);
		fil.saveFile(slotValues,
					 addressEditor->getText(),
					 portEditor->getText(),
					 midiInput->getSelectedId());
		lastFile = dlg.getResult().getFullPathName();
		lastDir = dlg.getResult().getParentDirectory();
	}
}

//------------------------------------------------------------------------------
void WindowContents::fillSlotValues(SlotValues& slot)
{
	int i;

	//Pads.
	for(i=0;i<3;++i)
	{
		slot.padValues[i].xAddress = pads[i]->getXAddress();
		slot.padValues[i].yAddress = pads[i]->getYAddress();
		slot.padValues[i].xMax = pads[i]->getXMax();
		slot.padValues[i].yMax = pads[i]->getYMax();
		slot.padValues[i].xMin = pads[i]->getXMin();
		slot.padValues[i].yMin = pads[i]->getYMin();
		slot.padValues[i].xVal = pads[i]->getXVal();
		slot.padValues[i].yVal = pads[i]->getYVal();
		slot.padValues[i].xCc = pads[i]->getXCc();
		slot.padValues[i].yCc = pads[i]->getYCc();
	}

	//Top sliders.
	for(i=0;i<3;++i)
	{
		slot.topSliderValues[i].address = topSliders[i]->getAddress();
		slot.topSliderValues[i].max = topSliders[i]->getMax();
		slot.topSliderValues[i].min = topSliders[i]->getMin();
		slot.topSliderValues[i].val = topSliders[i]->getSliderVal();
		slot.topSliderValues[i].cc = topSliders[i]->getCc();
	}

	//Bottom sliders.
	for(i=0;i<3;++i)
	{
		slot.bottomSliderValues[i].address = bottomSliders[i]->getAddress();
		slot.bottomSliderValues[i].max = bottomSliders[i]->getMax();
		slot.bottomSliderValues[i].min = bottomSliders[i]->getMin();
		slot.bottomSliderValues[i].val = bottomSliders[i]->getSliderVal();
		slot.bottomSliderValues[i].cc = bottomSliders[i]->getCc();
	}
}

//------------------------------------------------------------------------------
void WindowContents::updateWidgets(const SlotValues& slot)
{
	int i;
	float tempf;
	float tempMin, tempMax;

	//Pads
	for(i=0;i<3;++i)
	{
		pads[i]->setXAddress(slot.padValues[i].xAddress);
		pads[i]->setYAddress(slot.padValues[i].yAddress);
		pads[i]->setXMax(slot.padValues[i].xMax);
		pads[i]->setXMin(slot.padValues[i].xMin);
		pads[i]->setYMax(slot.padValues[i].yMax);
		pads[i]->setYMin(slot.padValues[i].yMin);
		pads[i]->setXVal(slot.padValues[i].xVal);
		pads[i]->setYVal(slot.padValues[i].yVal);
		pads[i]->setXCc(slot.padValues[i].xCc);
		pads[i]->setYCc(slot.padValues[i].yCc);
	}

	//Top Sliders
	for(i=0;i<3;++i)
	{
		tempMin = slot.topSliderValues[i].min.getFloatValue();
		tempMax = slot.topSliderValues[i].max.getFloatValue();

		topSliders[i]->setAddress(slot.topSliderValues[i].address);
		topSliders[i]->setMax(slot.topSliderValues[i].max);
		topSliders[i]->setMin(slot.topSliderValues[i].min);

		tempf = slot.topSliderValues[i].val;
		if(tempMin < tempMax)
			tempf -= tempMin;
		else
			tempf -= tempMax;
		tempf /= tempMax - tempMin;
		topSliders[i]->setValue(tempf);

		topSliders[i]->setCc(slot.topSliderValues[i].cc);
	}

	//Bottom Sliders
	for(i=0;i<3;++i)
	{
		tempMin = slot.bottomSliderValues[i].min.getFloatValue();
		tempMax = slot.bottomSliderValues[i].max.getFloatValue();

		bottomSliders[i]->setAddress(slot.bottomSliderValues[i].address);
		bottomSliders[i]->setMax(slot.bottomSliderValues[i].max);
		bottomSliders[i]->setMin(slot.bottomSliderValues[i].min);

		tempf = slot.bottomSliderValues[i].val;
		if(tempMin < tempMax)
			tempf -= tempMin;
		else
			tempf -= tempMax;
		tempf /= tempMax - tempMin;
		bottomSliders[i]->setValue(tempf);

		bottomSliders[i]->setCc(slot.bottomSliderValues[i].cc);
	}
}
